define CCF_2d()
%!%+
%\function{CCF_2d}
%\synopsis{calculates the cross correlation function of two 2d-arrays (e.g., images)}
%\usage{Double_Type CCF = CCF_2d(img1, img2, Integer_Type dx, Integer_Type dy);}
%\description
%     The 2d-arrays img1 and img2 from which the CCF is to be computed can
%     either be passed directly or by a reference to the arrays.
%     This function computes the correlation of \code{img1[y, x]} and \code{img2[y-dy, x-dx]},
%     where the indices (x,y) take all values such that (x,y) and (x-dx,y-dy) are contained:
%        \code{CCF = sum_{x,y} { (img1[y,x]-<img1>)/|img1| * (img2[y-dy,x-dx]-<img2>)/|img2| }}
%     The norm \code{|img| = sqrt{ sum_{x,y} img[y,x]^2 }} ensures that \code{CCF_2d(img, +-img, 0, 0) = +-1}.
%     for any image \code{img}.\n
%     If \code{dx} or \code{dy} are arrays, \code{CCF} will be a 2d array.
%\qualifiers{
%\qualifier{verbose}{turns on verbosity}
%\qualifier{savebest}{=&best: if \code{dx} or \code{dy} are arrays, the best combination will be saved in \code{best}.}
%}
%\seealso{CCF_1d}
%!%-
{
  variable img1, img2, dx, dy;
  switch(_NARGS)
  { case 4: (img1, img2, dx, dy) = (); }
  { help(_function_name()); return; }

%  if(typeof(img1)==Ref_Type) { img1 = @img1; };
%  if(typeof(img2)==Ref_Type) { img2 = @img2; };
  if(typeof(dx)==Array_Type || typeof(dy)==Array_Type)
  {
    dx = [dx];
    dy = [dy];
    variable CCF_best = 0;
    variable dx_best = NULL;
    variable dy_best = NULL;
    variable CCF = Double_Type[length(dy), length(dx)];
    variable idx, idy;
    _for idy (0, length(dy)-1, 1)
      _for idx (0, length(dx)-1, 1)
      {
        variable c = CCF_2d(img1, img2, dx[idx], dy[idy]);
	if(qualifier_exists("verbose"))
	  vmessage("dx=%d, dy=%d  =>  CCF = %.3f", int(dx[idx]), int(dy[idy]), c);
        if(abs(c)>abs(CCF_best))
	{ CCF_best = c;
	  dx_best = int(dx[idx]);
	  dy_best = int(dy[idy]);
	}
        CCF[idy, idx] = c;
      }
    vmessage("best: CCF(dx=%d, dy=%d) = %.3f", dx_best, dy_best, CCF_best);
    if(qualifier_exists("savebest"))
    { variable b = qualifier("savebest");
      @b = [dx_best, dy_best];
    }
    return CCF;
  }

  if(typeof(dx)!=Integer_Type)  dx = int(dx);
  if(typeof(dy)!=Integer_Type)  dy = int(dy);

  variable w = length(img1[0,*]);
  variable h = length(img1[*,0]);
  variable X1 = [int(_max(0,dx)) : w+int(_min(0,dx))-1];
  variable Y1 = [int(_max(0,dy)) : h+int(_min(0,dy))-1];

  img1 = img1[Y1   , X1   ];  img1 -= mean(img1); img1 /= sqrt(sum(img1^2));
  img2 = img2[Y1-dy, X1-dx];  img2 -= mean(img2); img2 /= sqrt(sum(img2^2));

  return sum(img1*img2);
}
