define sort_arrays()
%!%+
%\function{sort_arrays}
%\synopsis{}
%\usage{Integer_Type[] sort_arrays( Array_Type[] A1 [, A2 [, A3 ...]] );}
%\qualifiers{
%\qualifier{dir [=1]:}{1: Ascending sorting; -1: Descending sorting}
%\qualifier{method [='msort']:}{ Choose sorting method, 'msort' for merge-
%                       and 'qsort' for quick-sort.}
%}
%\description
%   Similar to 'array_sort' this funtion sorts arrays, but also sorts
%   multiple appearing array entries based on an addition given array
%   and so on. In other words this function allows to sort several arrays
%   (e.g., columns in a FITS-file) depending on the sorting of the previous
%   array.
%
%   This is especially useful if the arrays represent parameter combinations
%   and you want to perform a resorting. See example below.
%
%   NOTE THAT all arrays must have the same length!
%   
%\example
%   % Example 1:
%    a1 = [ 5,6,1,7,5,6,7 ];
%    a2 = [ 8,9,4,4,1,2,3 ];
%    i12= sort_arrays(a1,a2);
%    array_map( Void_Type, &vmessage, "%.2d %.2d", a1[i12], a2[i12] );
%    i21 = sort_arrays(a2,a1);
%    array_map( Void_Type, &vmessage, "%.2d %.2d", a1[i21], a2[i21] );
%
%   % Example 2: resorting parameter combinations:
%    par = struct{ a = [0:1:#3],
%                  b = ["hello","world"],
%                  c = [ 2001, 2012]
%                };
%    parcomb = get_par_combinations( par );
%    print(parcomb);
%
%    parcomb =  struct_array_2_struct_of_arrays(parcomb);
%    struct_filter( parcomb, sort_arrays(parcomb.b,parcomb.a,parcomb.c) );
%    print(struct_of_arrays_2_struct_array(parcomb));
%    
%\seealso{array_sort, get_par_combinations}
%!%-
{
  
  if( _NARGS == 0 ){
    help(_function_name);
    return NULL;
  }
  %% Get Arguments
  variable arg_list = __pop_list(_NARGS);

  %% Initialize sorting index array
  variable isort = array_sort( arg_list[0] ;; __qualifiers );
  variable len = length(isort);
  
  variable ii, istart, istopp;
  _for ii ( 0, _NARGS-2 ){ % loop over all given arrays
    istart = 0;
    while( istart < len ){
      % the ii-th array is sorted, [istart:istopp] gives
      % the global interval where the ii-th array has the same entry
      % not that it is crucial to use wherefirst here, which only works if the array is sorted!
      istopp = wherefirst( arg_list[ii][isort][[istart:]] != arg_list[ii][isort][istart] );
      if( _isnull(istopp) ){
	istopp = len-1;
      }else{
	istopp += (istart-1);
      }

      % now sort the ii+1th array in the [istart:istopp] range until the end of the array is hit
      isort[ [istart:istopp] ] = isort[ [istart:istopp] ][ array_sort( arg_list[ii+1][isort][[istart:istopp]] ;; __qualifiers ) ];
      istart = istopp+1;

      % when the ii+1th array is sorted go to the next array, i.e., the ii+1th array will be the ii-th array!
    }
  }

  return isort;      
}
