
require("gslinterp");

define barycen();

define barycen() {
%!%+
%\function{barycen}
%\synopsis{barycenter an event list, FITS extension, or FITS file}
%\usage{barytime=barycen(eventlist);}
%\altusage{barycen(String_Type fitsfile);}
%\altusage{barycen(FITS_File_Type fptr);}
%\qualifiers{
%\qualifier{debug}{if set, print debugging information}
%\qualifier{ephemeris}{pointer to or name of a FITS ephemeris file (default: DE430)}
%\qualifier{orbit}{mandatory, structure containing the orbit information, in
%                    geocentric ICRS coordinates and velocities in tags
%                    x,y,z,vx,vy,vz in units of m and m/s, and time in tag mjd}
%\qualifier{deltat}{if set and the routine barycenters an event list or times, return the
%                     difference between the barycentered time and the normal time}
%\qualifier{ra}{right ascension (ICRS) of the source [rad]}
%\qualifier{dec}{declination (ICRS) of the source [rad]}
%\qualifier{deg}{if set, ra/dec are in degrees}
%}
%\description
% This routine barycenters an event list, a FITS extension that is compatible
% with this routine, or all FITS extensions in a FITS file that are compatible
% with this routine.
%
% Caveat emptor: Right now the function assumes that all times are in TT and
% coordinates are in the ICRS system, i.e., it does not know that certain
% JPL ephemerides are NOT in this coordinate system.
%
%\seealso{jpl_eph}
%
%!%-
    variable arg1;

    arg1=();

    variable debug=qualifier("debug",0);
    %
    variable i,nrow,rr,dd,ra,de,oldcolnum,newcolnum;
    variable writefits=-1;

    %
    % initialize the ephemeris
    %
    variable eph, ephkey;
    if (qualifier_exists("ephemeris")) {
        eph=qualifier("ephemeris");
      if (typeof(eph) == Struct_Type){ % user may manually initialize orbit 
        ephkey="USER DEFINED";
      } else {
	ephkey=eph;
	eph=jpl_initeph(getenv("LHEA_DATA")+"/"+eph;nocache);
	
      }
    } else {
      ephkey="JPLEPH.430";
      eph=jpl_initeph(getenv("LHEA_DATA")+"/JPLEPH.430";nocache);
    }

    %
    % initialize the orbit
    % TO DO: also allow Earth coordinates
    %
    variable orbit=qualifier("orbit");

    variable status;
    variable hdu;

    if (typeof(arg1)==String_Type) {
        %
        % 1st argument is a string: file name
        % barycenter all extensions
        %
        variable qual=__qualifiers();
        if (not qualifier_exists("ephemeris")) {
            qual=struct_combine(qual,struct { ephemeris=eph} );
        }

        variable fptr=fits_open_file(arg1,"w");
        variable numhdu=fits_get_num_hdus(fptr);

        _for hdu(1,numhdu,1) {
            if (debug>0) {
                ()=printf("Extension number %i: ",hdu);
            }
            fits_movabs_hdu(fptr,hdu);

            variable extname, timeref;
            status=_fits_read_key(fptr,"EXTNAME",&extname,NULL);
            status=_fits_read_key(fptr,"TIMEREF",&timeref,NULL);
	  
            if (typeof(extname)==String_Type && (extname=="EVENTS" || extname=="RATE") ) {
	         if (typeof(timeref)==String_Type && (timeref!="SOLARSYSTEM")){
		  if (debug>0) {
                    ()=printf("%s - barycentering\n",extname);
		  }
		  barycen(fptr;; qual);
		 } else {
		   ()=printf("%s already barycentered! - skipping!\n",extname);
		 }
            } else {
                if (debug>0) {
                    if (typeof(extname)==String_Type) {
                        ()=printf("%s - skipping\n",extname);
                    } else {
                        ()=printf("skipping\n");
                    }
                }
            }
        }
        fits_close_file(fptr);
        return;
    }

    if (typeof(arg1)==Fits_File_Type) {
        %
        % we are working on a Fits file extension
        %
        writefits=1;
        fptr=arg1;
    
        variable evthdr=fits_read_header(fptr);

        variable colnames=qualifier("colnames",["TIME","TSTART","TSTOP"]);
        variable newnames=qualifier("newnames",["TBARY","TSTART","TSTOP"]);

        if (length(colnames)!=length(newnames)) {
            throw UsageError,sprintf("%s: Usage error: colnames and newnames must have same length\n",_function_name());
        }
        
        oldcolnum={};
        newcolnum={};
    
        % check existence of possible time columns,
        % generate barycentric columns if required
        _for i(0,length(colnames)-1,1) {
            if (fits_binary_table_column_exists(fptr,colnames[i])) {
                variable timecol=fits_get_colnum(fptr,colnames[i]);
                list_append(oldcolnum,timecol);
                if (colnames[i]!=newnames[i]) {
                    variable barycol=timecol+1;
                    list_append(newcolnum,barycol);
                    status=_fits_insert_cols(fptr,barycol,[newnames[i]],["D"]);
                }
            }
        }
    
        % number of events
        nrow=evthdr.NAXIS2;

        % source position
        % rr,dd will be in deg, ra,de in rad
        rr=qualifier("ra",evthdr.RA_OBJ);
        dd=qualifier("dec",evthdr.DEC_OBJ);

        if (qualifier_exists("ra")) {
            if (qualifier_exists("deg")) {
                ra=rr*PI/180.;
            } else {
                ra=rr;
                rr*=180./PI; % convert to degrees for FITS header
            }
        } else {
            ra=rr*PI/180.;
        }

        if (qualifier_exists("dec")) {
            if (qualifier_exists("deg")) {
                de=dd*PI/180.;
            } else {
                de=dd;
                dd*=180./PI;
            }
        } else {
            de=dd*PI/180.;
        }
    }

    if (typeof(arg1[0])==Double_Type) {
        %
        % direct barycentering a time value or array
        %
        writefits=0;
        
        if (qualifier_exists("ra")) {
            ra=qualifier("ra");
        } else {
            throw UsageError,sprintf("%s: Usage error: need ra to barycenter a time array\n",_function_name());
        }

        if (qualifier_exists("dec")) {
            de=qualifier("dec");
        } else {
            throw UsageError,sprintf("%s: Usage error: need dec to barycenter a time array\n",_function_name());
        }

        if (qualifier_exists("deg")) {
            ra*=PI/180.;
            de*=PI/180.;
        }

        % ra, de are in rad
        % rr, dd are in deg        
        rr=ra*180./PI; 
        dd=de*180./PI;

        nrow=length(arg1);
        oldcolnum=0;

        variable tnew=Double_Type(length(arg1));
    }

    if (writefits==-1) {
        throw UsageError,sprintf("%s: first argument must be Double array or FITS file pointer\n",_function_name());
    }
    
    % unit vector towards the source
    variable srcdir=struct {x=cos(de)*cos(ra),y=cos(de)*sin(ra),z=sin(de)};

    %
    % now loop over all times
    %

    % speed of light
    variable c=299792458.; % m/s
    variable GMsun=132712440041.939400e9; % m^3/s^2 (per DE430)

    variable row,col;
    _for row (1,nrow,1) {
        if (debug>0) {
            if (row mod 10000==1) {
                ()=printf("Processing row %7i of %7i (%5.1f%% done)\n",row,nrow,(100.*row)/nrow);
            }
        }

        variable evt_tt;

        % loop over the columns we need to barycenter
        % note: if we barycenter an array, we set oldcolnum to 0 above, i.e.,
        % the loop is only executed once
        _for col (0,length(oldcolnum)-1,1) {
            if (writefits==1) {
                variable evt_time;
                status=_fits_read_col(fptr,timecol,row,1,&evt_time);
                evt_time=evt_time[0];
        
                % convert event time to mjd (in TT)
                evt_tt=(evthdr.TIMEZERO+evt_time)/86400. + evthdr.MJDREF;
            } else {
                evt_tt=arg1[row-1];
            }

            variable earthbary=jpl_eph_vec(evt_tt,"Earth";mjd,equatorial,mks,barycentric,ephemeris=eph);
            variable sunbary=jpl_eph_vec(evt_tt,"Sun";mjd,equatorial,mks,barycentric,ephemeris=eph);

            %
            % satellite position wrt center of Earth 
            variable satearth=struct {
                x=interpol(evt_tt,orbit.mjd,orbit.x),
                y=interpol(evt_tt,orbit.mjd,orbit.y),
                z=interpol(evt_tt,orbit.mjd,orbit.z),
                vx=interpol(evt_tt,orbit.mjd,orbit.vx),
                vy=interpol(evt_tt,orbit.mjd,orbit.vy),
                vz=interpol(evt_tt,orbit.mjd,orbit.vz)
            };

            % note: in principle it does not make sense to measure the SAT time in TT,
            % it can well be that in this step we are doing double counting. This might
	    % introduce an error at the few mus(? - CHECK) level

            % barycentric position of the satellite
            variable satbary=struct {
               x=satearth.x+earthbary.x,
               y=satearth.y+earthbary.y,
               z=satearth.z+earthbary.z
	    };
 
            % Roemer delay (s)
            variable troemer=-(satbary.x*srcdir.x + satbary.y*srcdir.y + satbary.z*srcdir.z)/c;
            
            % Shapiro delay due to Sun (s), see Lyne & Graham-Smith, Pulsar Astronomy, Eq. 5.12
	    % vector sun->satellite
	    variable Rsun=struct {
	        x=satbary.x-sunbary.x,
	        y=satbary.y-sunbary.y,
	        z=satbary.z-sunbary.z
	    };
	    variable dsun=sqrt(Rsun.x*Rsun.x+Rsun.y*Rsun.y+Rsun.z*Rsun.z);
	    variable rang=Rsun.x*srcdir.x+Rsun.y*srcdir.y+Rsun.z*srcdir.z;
	    variable tshapiro=-2.*GMsun*log(1+rang/dsun)/c/c/c;

            variable t2b=-tt2tdb(evt_tt;mjd,deltat); % tt-tdb, change sign because tt2tdb returns TDB-TT

            % time dilatation (Einstein delay) needed for conversion TCG to TCB and accounts for
	    % time shift between surface of Earth and geocenter
	    % See Petit & Luzum (2010), IERS Technical Note No 36, Eq. 10.5

	    variable tdil=-(satearth.x*earthbary.vx+satearth.y*earthbary.vy+satearth.z*earthbary.vz)/c/c;

            if (debug>0) {
	        if (row mod 10000==1) {
	            ()=printf("   roemer: %12.3f ms shapiro: %7.3g ms\n",troemer*1e3,tshapiro*1e3);
	            ()=printf("   t2b   : %12.3f ms tdil   : %7.3f ms\n",t2b*1e3,tdil*1e3);
	        }
            }
	    variable deltat=t2b+tdil+troemer+tshapiro;
        
            % now store barycentric value
            if (writefits==1) {
                status=_fits_write_col(fptr,newcolnum[col],row,1,evt_time-deltat);
            } else {
                if (qualifier_exists("deltat")) {
                    % return tt-tbary
                    tnew[row-1]=deltat;
                } else {
                    tnew[row-1]=evt_time-deltat;
                }
            }
        }
    } % end of event loop
        
    if (debug>0) {
        ()=printf("Processing row %7i of %7i (%5.1f%% done)\n",nrow,nrow,100.);
    }

    if (writefits==1) {
        status=_fits_update_key(fptr,"TIMEREF","SOLARSYSTEM","reference system");
        status=_fits_update_key(fptr,"PLEPHEM",ephkey,"ephemeris used");
        status=_fits_update_key(fptr,"RA_TDB",rr,"[deg] RA of TDB correction");
        status=_fits_update_key(fptr,"DEC_TDB",dd,"[deg] DEC of TDB correction");
        _for i(0,length(newcolnum)-1,1) {
            if (oldcolnum[i]!=newcolnum[i]) {
                status=_fits_update_key(fptr,sprintf("TUNIT%i",newcolnum[i]),"s",sprintf("Unit of %s column",newnames[i]));
            }
        }
        status=_fits_write_chksum(fptr);
    }

    if (writefits==0) {
        return tnew;
    }
}

