define cosmo_param() %,omega_m, omega_lambda, omega_k, q0
%!%+
%\function{cosmo_param}
%\synopsis{Derive full set of cosmological density parameters from a partial set}
%\usage{Struct_Type P = cosmo_param();}
%\qualifiers{
%\qualifier{omega_m}{value of the normalized matter density (Omega_M)}
%\qualifier{omega_lambda}{value of the normalized cosmological constant (Omega_Lambda)}
%\qualifier{omega_k}{value of the the normalized curvature term (Omega_k)}
%\qualifier{q0}{value of the deceleration parameter (q0)}
%}
%\description
%    This procedure is called by lumdist to allow the user a choice
%    in defining any two of four cosmological density parameters.
%    Given any two of the four input parameters this  program will derive
%    the remaining two.
%    Omega_M      - Normalized matter energy density, non-negative numeric scalar
%    Omega_Lambda - Normalized cosmological constant, numeric scalar
%    Omega_k      - Normalized curvature parameter, numeric scalar.
%                   This is zero for a flat universe
%    q0           - Deceleration parameter, numeric scalar = -R*(R'')/(R')^2
%                   = 0.5*Omega_m - Omega_lambda
%    Here "normalized" means divided by the closure density so that
%    Omega_m + Omega_lambda + Omega_k = 1.    For a more
%    precise definition see Carroll, Press, & Turner (1992, ArAA, 30, 499).
%
%    If less than two parameters are defined, this procedure sets default
%    values of Omega_k=0 (flat space), Omega_lambda = 0.7, Omega_m = 0.3
%    and q0 = -0.55
%
%    If more than two parameters are defined upon input (overspecification), 
%    then the first two defined parameters in the ordered list Omega_m, 
%    Omega_lambda, Omega_k, q0 are used to define the cosmology.
%    
%    (This procedure is translated from the IDL COSMO_PARAM function.)
%
%\example
%    variable default_set = cosmo_param();
%    print (default_set);
%
%    variable my_set = cosmo_param(; omega_k = 0.1, omega_lambda = 0.65);
%    print (my_set);
%\seealso{lumdist}
%!%-
{
  variable omega_m      = qualifier("omega_m"      ,NULL);
  variable omega_lambda = qualifier("omega_lambda" ,NULL);
  variable omega_k      = qualifier("omega_k"      ,NULL);
  variable q0           = qualifier("q0"           ,NULL);
  
  % check if parameters are set:
  variable Nk      = (omega_k      != NULL);
  variable Nlambda = (omega_lambda != NULL);
  variable Nomega  = (omega_m      != NULL);
  variable Nq0     = (q0           != NULL);
  
%  if ( Nk + Nlambda + Nomega + Nq0 < 2) % less than 2 params are set
%  {
%    message("Two parameters have to be set.");
%    return;
%  }
  if ( Nk + Nlambda + Nomega + Nq0 > 2)
    vmessage("WARNING -- %s -- More than 2 parameters are set!", _function_name);
  
  % Check which two parameters are defined, and then determine the other two
  if (Nlambda and Nq0)
  {
    if (     Nk == 0)      omega_k = 1. - (2.*q0 + 3.*omega_lambda);
    if ( Nomega == 0)      omega_m = 1.- (omega_lambda + omega_k);
  }
  if (Nk and Nq0)
  {
    if ( Nomega == 0)      omega_m = (1 + q0 - omega_k)*2./3.;
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
  }
  if (Nomega and Nq0)
  {
    if (     Nk == 0)      omega_k = 1. + q0 - 3.*omega_m/2.;
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
  }
  if (Nlambda and Nk)
  {
    if ( Nomega == 0)      omega_m = 1.- (omega_lambda + omega_k);
    if (    Nq0 == 0)           q0 = (1 - (omega_k + 3.*omega_lambda))/2.;
  }
  if (Nomega and Nk)
  {
    if (Nlambda == 0) omega_lambda = 1. - (omega_m + omega_k);
    if (    Nq0 == 0)           q0 = -1. + omega_k + 3*omega_m/2;
  }
 if (Nomega and Nlambda)
  {
    if (     Nk == 0)      omega_k = 1. - (omega_m + omega_lambda);
    if (    Nq0 == 0)           q0 = omega_m/2.  - omega_lambda;
  }
  
  %Set default values
  if  (omega_k      == NULL) omega_k      = 0.;       %Default is flat space
  if  (omega_lambda == NULL) omega_lambda = 0.7;
  if  (omega_m      == NULL) omega_m      = 1 - omega_lambda;
  if  (q0           == NULL) q0           = (1 - (omega_k + 3*omega_lambda))*0.5;
  
  return struct{
    omega_k      = omega_k,
    omega_lambda = omega_lambda,
    omega_m      = omega_m,
    q0           = q0
  };
}
