define dms2deg(d,m,s) {
%!%+
%\function{dms2deg}
%\synopsis{Convert angle in degree, minute, seconds to floating points degrees or radian}
%\usage{degree = dms2deg(d,m,s);}
%\qualifiers{
%\qualifier{hours}{If set, the coordinate is given in hour, minutes, sec (RA)}
%\qualifier{radian}{If set, return angle in radians, not degrees}
%}
%\description
%This is a convenience routine to convert astronomical coordinates given
%in degrees, minutes, and seconds into a floating point number.
%The sign of the highest non-zero argument decides the sign of the
%returned floating point number. In other words: The routine tries to 
%be intelligent for negative declinations in the 0>=dec>-1 strip. 
%For example, if the declination is -0d 14' 30.2", call this routine
%as dms2deg(0,-14,30.2).
%
%This routine is array safe (as long as d, m, s are arrays of equal length).
%\example
% % Convert the coordinates 
% % alpha=19h 49m 35.49s and dec=-30d 12' 31.8"
% variable ra=hms2deg(19,49,35.49);
% variable dec=dms2deg(-30,12,31.8);
%
%\seealso{hms2deg}
%!%-
   variable sgn;
   if (typeof(d)==Array_Type || typeof(m)==Array_Type || typeof(s)==Array_Type) {
       if (length(d)!=length(m) || length(m)!=length(s)) {
           throw UsageError("%s: d,m,s arguments must be of equal length\n",_function_name());
       }
       sgn=Double_Type[length(d)];
       variable ndx,cmplndx;
       ndx=where(d!=0.,&cmplndx);
       if (length(ndx)>0) {
           % everything not close to the equator
           sgn[ndx]=sign(d[ndx]);
       }
       if (length(cmplndx)>0) {
           % equator
           variable ndx2,cmplndx2;
           ndx2=where(m[cmplndx]!=0.,&cmplndx2);
           if (length(ndx2)>0) {
               sgn[cmplndx[ndx2]]=sign(m[cmplndx[ndx2]]);
           }
           if (length(cmplndx2)>0) {
               % d, m are 0
               variable ndx3,cmplndx3;
               ndx3=where(s[cmplndx[cmplndx2]]<0.,&cmplndx3);
               if (length(ndx3)!=0) {
                   sgn[cmplndx[cmplndx2[ndx3]]]=-1.;
               }
               if (length(cmplndx3)>0) {
                   sgn[cmplndx[cmplndx2[cmplndx3]]]=+1.;
               }
           }
       }
   } else {
       % case that d, m, s are not arrays
       if (d!=0.) {
           sgn=sign(d);
       } else {
           if (m!=0.) {
               sgn=sign(m);
           } else {
               sgn=sign(s);
           }
       }
   }

   variable ret = sgn*(abs(d)+(abs(m)+abs(s)/60.)/60.);

   if (qualifier_exists("hours")) {
       ret*=15.;
   }

   if (qualifier_exists("radian")) {
       ret*=PI/180.;
   }

   return ret;
}

define hms2deg(h,m,s) {
%!%+
%\function{hms2deg}
%\synopsis{Convert angle in hour, minute, seconds to degrees or radian}
%\usage{degree = hms2deg(h,m,s);}
%\qualifiers{
%\qualifier{radian}{If set, return angle in radians, not degrees}
%}
%\description
%This is a convenience routine to convert astronomical coordinates given
%in hours, minutes, and seconds into a floating point number.
%
%The routine is equivalent to calling dms2deg with the hours qualifier.
%
%This routine is array safe (as long as h, m, s are arrays of equal length).
%
%\seealso{dms2deg}
%!%-
    if (qualifier_exists("radian")) {
	return dms2deg(h,m,s;hours,radian);
    } else {
	return dms2deg(h,m,s;hours);
    }
}
