%
% Implement the IAU 2000A Nutation theory
%

%
% private variables to cache the nutation theories
%
private variable nutation2000a_ls;
private variable nutation2000a_pl;
private variable nutation2000k_ls;
private variable nutation2000k_pl;
private variable nutation2000b_ls;

define nutation_angles();

define nutation_angles()
%!%+
%\function{nutation_angles}
%\synopsis{calculate the nutation in longitude and obliquity for a given date}
%\usage{ (dpsi,deps)=nutation_angles(JD;qualifiers)}
%\qualifiers{
%   \qualifier{JD_frac}{fractional part of the Julian Date (for highest precision)}
%   \qualifier{theory}{nutation theory to use. Possible string values are
%           IAU2000A - full IAU2000A theory, i.e., an implementation of the
%                      theory of Matthews et al. (2002, J. Geophys. Res, 107, 2068)
%           IAU2000B - truncated IAU2000A theory per McCarthy & Luzum (2003,
%                      Cel. Mech. Dyn. Astron, 85, 37); accuracy is
%                      1 mas in 1995-2050
%           NU2000K  - truncated IAU2000A theory from Kaplan (2009, USNO
%                      circular 181); accuracy is 0.1mas in 1700-2300
%                      [the default]}
%   \qualifier{mjd}{the time argument is in MJD, not in JD}
%   \qualifier{JD_frac}{fractional part of the time argument. Use this if
%                     highest precision is needed. The nutation angles are then
%                     returned for the date JD+JD_frac.}
%   \qualifier{deg}{return nutation angles in degrees}
%   \qualifier{arcsec}{return nutation angles in arcseconds}
%   \qualifier{mas}{return nutation angles in milliarcseconds}
%   \qualifier{iaufile}{FITS file containing the theory coefficients
%                   (usually found in ISISSCRIPTS_REFPATH)}
% 
%}
%\description
% This routine calculates the nutation angles for the forced nutation
% of the non-rigid earth according to one of the nutation theories described
% above for date given by the JD-argument (which is interpreted as TT).
% For most applications, usage of the NU2000K or IAU2000B routines
% is fully sufficient.
%
% This routine is array safe (but JD_frac MUST be a scalar in this case!),
% the angles returned are in radians unless one of the deg, arcsec, or mas
% keywords is given.
%
% The routine is based on the NOVAS 3.1 C-code (Bangert et al., 2011, see
% http://aa.usno.navy.mil/software/novas/novas_info.php )
%
%\seealso{precess}
%!%-
{
    variable JD;
    switch(_NARGS)
    { case 1 : JD=(); }
    { help(_function_name()); return;}

    %
    % if called with an array, call this routine for each time argument
    %
    if (Array_Type==typeof(JD)) {
	variable i;
	variable dpsi=Double_Type[length(JD)];
	variable deps=Double_Type[length(JD)];
	_for i(0,length(JD)-1,1) {
	    (dpsi[i],deps[i])=nutation_angles(JD[i];;__qualifiers());
	}
	return dpsi,deps;
    }
    
    %
    % read FITS file?
    %
    ifnot (__is_initialized(&nutation2000a_ls)) {
	variable searchpath=getenv("ISISSCRIPTS_REFPATH");
	if (searchpath==NULL and qualifier_exists("iaufile")==0) {
          throw UsageError,sprintf("%s: iaufile qualifier not given and ISISSCRIPTS_REFPATH not found\n",_function_name());
      }
      if (searchpath==NULL) {
          searchpath=".";
      }
      variable nutfi=qualifier("iaufile",searchpath+"/nutation_iau2000.fits.gz");
      % the following opens the file multiple times
      nutation2000a_ls=fits_read_table(nutfi+"[IAU2000A_LS]");
      nutation2000a_pl=fits_read_table(nutfi+"[IAU2000A_PL]");
      nutation2000b_ls=fits_read_table(nutfi+"[IAU2000B_LS]");
      nutation2000k_ls=fits_read_table(nutfi+"[IAU2000K_LS]");
      nutation2000k_pl=fits_read_table(nutfi+"[IAU2000K_PL]");
    }


    % get the theory coefficients
    variable theory_ls, theory_pl;
    variable theory=qualifier("theory","NU2000K");
    switch (theory)
    { case "IAU2000A":
	theory_ls=nutation2000a_ls;
	theory_pl=nutation2000a_pl;
    }	
    { case "IAU2000B":
	theory_ls=nutation2000b_ls;
	theory_pl=NULL;
    }
    { case "NU2000K":
	theory_ls=nutation2000k_ls;
	theory_pl=nutation2000k_pl;
    }
    {throw UsageError,sprintf("%s: Unknown theory for nutation.\n",_function_name()); }

    variable T;
    if (qualifier_exists("mjd")) {
	T=((JD-51544.5)+qualifier("JD_frac",0.))/36525.0;
    } else {
	T=((JD-2451545.)+qualifier("JD_frac",0.))/36525.0;
    }

    % conversion parameters
    variable asec360=3600.*360.;
    variable asec2rad=PI/648000.;

    variable L, Lprime, F, D, Om;
    if (theory=="IAU2000B") {
	% Delaunay angles from Simon et al. (1994)
	% (used in IAU 2000B)


	% Mean anomaly of the Moon
	L = ((485868.249036 + T * 1717915923.2178) mod asec360) * asec2rad;

	% Mean anomaly of the Sun
	Lprime = ((1287104.79305 + T * 129596581.0481) mod asec360) * asec2rad;

	% Mean argument of the latitude of the Moon
	F = ((335779.526232 + T * 1739527262.8478) mod asec360) * asec2rad;

	% Mean elongation of the Moon from the Sun.
	D = ((1072260.70369 + T * 1602961601.2090) mod asec360) * asec2rad;

	% Mean longitude of the ascending node of the Moon.
	Om = ((450160.398036 - T * 6962890.5431) mod asec360) * asec2rad;
    } else {	
	% Fundamental arguments from the Simon et al. (1994) lunar theory
	% full polynomials are used for IAU 2000A and NU 2000K

	% Mean anomaly of the Moon
        L = ((485868.249036 +
                   T * (1717915923.2178 +
                   T * (        31.8792 +
                   T * (         0.051635 +
                   T * (       - 0.00024470))))) mod asec360) * asec2rad;

	% Mean anomaly of the Sun
        Lprime= ((1287104.79305 +
                       T * ( 129596581.0481 +
                       T * (       - 0.5532 +
                       T * (         0.000136 +
                       T * (       - 0.00001149))))) mod asec360) * asec2rad;

	% Mean argument of the latitude of the Moon
        F = ((335779.526232 +
                   T * (1739527262.8478 +
                   T * (      - 12.7512 +
                   T * (      -  0.001037 +
                   T * (         0.00000417))))) mod asec360) * asec2rad;

	% Mean elongation of the moon from the Sun
	D = ((1072260.70369 +
                   T * (1602961601.2090 +
                   T * (       - 6.3706 +
                   T * (         0.006593 +
                   T * (       - 0.00003169))))) mod asec360) * asec2rad;

	% Mean longitude of the Moon's ascending node
	Om= ((450160.398036 +
                   T * ( - 6962890.5431 +
                   T * (         7.4722 +
                   T * (         0.007702 +
                   T * (       - 0.00005939))))) mod asec360) * asec2rad;
    } 
	
    %
    % calculate the luni-solar nutation
    %
    variable twopi=2.*PI;
    
    variable arg=(theory_ls.l*L+theory_ls.lprime*Lprime+
                 theory_ls.f*F+theory_ls.d*D+theory_ls.om*Om) mod twopi;
    variable sc,cc;
    (sc,cc)=sincos(arg);

    %
    % should probably do a reverse sort first
    %
    variable dp_ls=sum((theory_ls.l_sin+T*theory_ls.l_tsin)*sc+theory_ls.l_cos*cc);
    variable de_ls=sum((theory_ls.e_cos+T*theory_ls.e_tcos)*cc+theory_ls.e_sin*sc);
    
    %
    % planetary nutation
    %
    variable dp_pl,de_pl;
    if (NULL!=theory_pl) {
	variable al,alsu,af,ad,aom,apa,alme,alve,alea,alma,alju,alsa,alur,alne;

	if (theory=="IAU2000A") {
	    %
	    % anomalies and longitudes for the IAU2000A theory
	    %
	    
	    % mean anomaly of the Moon 
	    al = (2.35555598 + 8328.6914269554 * T) mod twopi;

	    % Mean anomaly of the Sun
	    alsu=(6.24006013 + 628.301955 * T) mod twopi;

	    % Mean argument of the latitude of the Moon
	    af = (1.627905234 + 8433.466158131 * T) mod twopi;

	    % Mean elongation of the Moon from the Sun
	    ad = (5.198466741 + 7771.3771468121 * T) mod twopi;

	    % Mean longitude of the ascending node of the Moon.
	    aom = (2.18243920 - 33.757045 * T) mod twopi;

	    % General accumulated precession in longitude.
	    apa = (0.02438175 + 0.00000538691 * T) * T;

	    % Planetary longitudes, Mercury through Neptune (Souchay et al. 1999).
	    alme = (4.402608842 + 2608.7903141574 * T) mod twopi;
	    alve = (3.176146697 + 1021.3285546211 * T) mod twopi;
	    alea = (1.753470314 +  628.3075849991 * T) mod twopi;
	    alma = (6.203480913 +  334.0612426700 * T) mod twopi;
	    alju = (0.599546497 +   52.9690962641 * T) mod twopi;
	    alsa = (0.874016757 +   21.3299104960 * T) mod twopi;
	    alur = (5.481293871 +    7.4781598567 * T) mod twopi;
	    alne = (5.321159000 +    3.8127774000 * T) mod twopi;
	} else {
	    %
	    % Terms for the IAU 2000K theory
	    %

	    %
	    % lunar terms (same as for IAU 2000A terms above)
	    %
	    al=L;
	    alsu=Lprime;
	    af=F;
	    ad=D;
	    aom=Om;
	    
	    % planetary longitudes from mercury through neptune wrt
	    % dynamical ecliptic and equinox of J2000, higher order
	    % terms omitted (Simon et al., 1994, Sect. 5.8.1-5.8.8)
	    alme = (4.402608842461 + 2608.790314157421 * T) mod twopi;
	    alve = (3.176146696956 + 1021.328554621099 * T) mod twopi;
	    alea = (1.753470459496 +  628.307584999142 * T) mod twopi;
	    alma = (6.203476112911 +  334.061242669982 * T) mod twopi;
	    alju = (0.599547105074 +   52.969096264064 * T) mod twopi;
	    alsa = (0.874016284019 +   21.329910496032 * T) mod twopi;
	    alur = (5.481293871537 +    7.478159856729 * T) mod twopi;
	    alne = (5.311886286677 +    3.813303563778 * T) mod twopi;

	    % General precession in longitude (Simon et al. 1994)
	    apa = (0.024380407358 + 0.000005391235 * T) * T;
	}

	% theory arguments 
	arg = (al*theory_pl.l   + alsu*theory_pl.lprime+af*theory_pl.f+
               ad*theory_pl.d   + aom*theory_pl.om+
               alme*theory_pl.me+alve*theory_pl.ve+
               alea*theory_pl.e +alma*theory_pl.ma+
               alju*theory_pl.ju+alsa*theory_pl.sa+
               alur*theory_pl.ur+alne*theory_pl.ne+
               apa*theory_pl.pre) mod twopi;
	(sc,cc)=sincos(arg);

	% should sort the array before summing
	dp_pl=sum(theory_pl.l_sin*sc+theory_pl.l_cos*cc);
	de_pl=sum(theory_pl.e_sin*sc+theory_pl.e_cos*cc);
    } else {
	%
	% fixed terms accounting for omission of long-period planetary
	% terms in the IAU 2000B theory
	%
	dp_pl=-1350.0; % 1e-7 arcsec
	de_pl= 3880.0; % 1e-7 arcsec
    }

    variable dp,de;
    if (qualifier_exists("deg")) {
	dp=(dp_ls+dp_pl)*1e-7*3600.;
	de=(de_ls+de_pl)*1e-7*3600.;
    } else {
	if (qualifier_exists("arcsec")) {
	    dp=(dp_ls+dp_pl)*1e-7;
	    de=(de_ls+de_pl)*1e-7;
	} else {
	    if (qualifier_exists("mas")) {
		dp=(dp_ls+dp_pl)*1e-1;
		de=(de_ls+de_pl)*1e-1;
	    } else {
		% radians
		dp=(dp_ls+dp_pl)*1e-7*asec2rad;
		de=(de_ls+de_pl)*1e-7*asec2rad;
	    }
	}
    }
    return (dp,de);
}
