define period_search_via_string_length() %{{{
%!%+
%\function{period_search_via_string_length}
%\synopsis{A period-finding method for sparse randomly spaced observations}
%\usage{(Double_Type p[], SL[]) = period_search_via_string_length(Double_Type t[], s[], pmin, pmax; qualifiers)}
%\description
%    This function may be useful to determine the period of a variable
%    signal 's' observed at a relatively small number 'N' of randomly
%    spaced observations at times 't' possibly taken many periods apart.
%
%    To identify candidate periods in the given range between 'pmin' and
%    'pmax', string-lengths 'SL' for trial periods 'p' (sampled in equal
%    frequency steps, see qualifier "delta_phi_max") are computed and
%    tabulated. The trial period giving the lowest string-length results
%    in the most smooth-looking curve and is thus the most likely period.
%\notes
%    The underlying idea is to minimize the so-called "string-length" in
%    a phase diagram. The phase diagram for a trial period 'p' is created
%    by computing the phase 'phi' according to
%      phi = (t-min(t))/p mod 1.
%    The string-length 'SL' is then simply the sum of the lengths of line
%    segments joining successive points (phi_i, s_i) in a phase diagram:
%      SL = sum ( ((s_i-s_i-1)^2+(phi_i-phi_i-1)^2)^(1/2) )
%         + ((s_1-s_N)^2+(phi_1-phi_N+1)^2)^(1/2).
%    To get rid of different scales caused by different units in ordinate
%    (e.g., km/s or magnitude) and abscissa (unit = trial period), the
%    observations are scaled so that
%      s' = 0.5*(s-min(s))/(max(s)-min(s))-0.25
%    in order to give equal weight to measures and phases (for details,
%    see Dworetsky 1983, MNRAS, 203, 917).
%
%    For reference: the string-length of a sinusoidal string is 1.4637.
%\qualifiers{
%\qualifier{delta_phi_max [=0.01]}{: Trial periods 'p' are sampled in equal
%      frequency steps of
%        delta_freq = delta_phi_max/(max(t)-min(t)).
%      This ensures that phase errors are always less than 'delta_phi_max'
%      since
%        phi = (t-min(t))/p mod 1 = (t-min(t))*freq mod 1
%        -> delta_phi = (t-min(t))*delta_freq
%        -> delta_phi_max = (max(t)-min(t))*delta_freq.}
%}
%\example
%    t = 1000*urand(50); % time of observations
%    s = sin(t)+0.05*grand(length(t)); % signal of observations; true period is 2*PI
%    connect_points(0); point_style(2); plot(t, s);
%    (p, SL) = period_search_via_string_length(t, s, 1, 10; delta_phi_max=0.01);
%    connect_points(-1); point_style(-1); plot(p, SL);
%    ind = where(SL==min(SL)); SL[ind[0]]; p[ind[0]]; % index, value, and period of minimum string-length
%    connect_points(0); point_style(2); plot((t-min(t))/p[ind[0]] mod 1, s); % phase diagram
%\seealso{}
%!%-
{
  if(_NARGS!=4)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable t, s, pmin, pmax;
  (t, s, pmin, pmax) = ();
  if(length(t)!=length(s))
    throw UsageError, sprintf("Usage error in '%s': The first two arguments have to be arrays of the same length.", _function_name());
  if(pmin>=pmax)
    throw UsageError, sprintf("Usage error in '%s': The minimum period has to be lower than the maximum period.", _function_name());
  % ensure that times start at 0 and are positive:
  t -= min(t);
  % scale signal according to Eq. (3) in Dworetsky 1983, MNRAS, 203, 917:
  variable mis = min(s), mas = max(s);
  s = 0.5*(s-mis)/(mas-mis)-0.25;
  % create array with trial periods:
  variable delta_freq = qualifier("delta_phi_max", 0.01)/max(t); % min(t)=0, see above
  variable len = (1./pmin-1./pmax)/delta_freq;
  if(len>100000000) % to avoid runtime problems because of huge random access memory usage; Double_Type[100000000] consumes roughly 1GB
    throw LimitExceededError, sprintf("Usage error in '%s': Random access memory usage too large. Choose a smaller interval for the trial periods.", _function_name());
  len = nint(len);
  variable periods = 1./[1./pmax:1./pmin:#len];
  variable string_length = Double_Type[len];
  % phase data according to trial period and compute string length:
  variable i;
  _for i(0, len-1, 1)
  {
    % phase data according to trial period:
    variable phase = t/periods[i] mod 1;
    % sort in ascending order with phase:
    variable ind = array_sort(phase); phase = phase[ind];
    variable phased_s = s[ind];
    % compute string-length:
    string_length[i] = sum(sqrt((phased_s[[1:]]-phased_s[[:-2]])^2+(phase[[1:]]-phase[[:-2]])^2))+sqrt((phased_s[0]-phased_s[-1])^2+(phase[0]-phase[-1]+1)^2);
  }
  return (periods, string_length);
}%}}}
