define polytropic_standard_model() %{{{
%!%+
%\function{polytropic_standard_model}
%\synopsis{Compute the structure of a polytropic standard model}
%\usage{Struct_Type s = polytropic_standard_model(Double_Type M, R, X, Y)}
%\description
%    Based on the stellar mass 'M' (in solar masses), stellar
%    radius 'R' (in solar radii), hydrogen mass fraction 'X',
%    and helium mass fraction 'Y', this function computes the
%    structure of a polytropic standard model. The output
%    structure 's' contains the fields "r" (radial coordinate
%    in solar radii), "rho" (density in g per cm^3), "P"
%    (pressure in dynes per cm^2), and "T" (temperature in K),
%    i.e., the structure of the polytrope.
%\notes
%    Polytropes are gaseous spheres in hydrostatic equilibrium
%    which obey a polytropic equation of state
%      P = K rho^((n+1)/n)
%    where 'P' is the pressure, 'rho' the density, 'K' the
%    polytropic constant, and 'n' the polytropic index. The
%    so-called polytropic standard model is a polytropic model
%    of index n=3. Its structure has a surprisingly good
%    resemblence to those of main sequence stars and is, thus,
%    used as a starting point to study the internal structure
%    of stars. For details, see, e.g., Chapter 2.4 in the book
%    "Principles of Stellar Evolution and Nucleosynthesis" by
%    D.D. Clayton.
%\example
%    s = polytropic_standard_model(1, 1, 0.73, 0.26);
%\seealso{solve_Lane_Emden_equation, solve_Eddington_quartic_equation}
%!%-
{
  if(_NARGS!=4)
  {
    _pop_n(_NARGS); % remove function arguments from stack
    help(_function_name());
    throw UsageError, sprintf("Usage error in '%s': wrong number of arguments.", _function_name());
  }
  variable M, R, X, Y;
  (M, R, X, Y) = ();
  M *= 1.; R *= 1.; % cast to Double_Type
  variable mu = 2./(1.+3.*X+0.5*Y); % mean molecular weight, see Equation (2-15) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable beta = solve_Eddington_quartic_equation(M, mu);
  variable K = ((8.314510e7/mu)^4*3/7.5646e-15*(1.-beta)/beta^4)^(1./3.); % polytropic constant of a standard model, see Equation (2-302) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable Pc = 1.24e17*M^2/R^4; % central pressure of a polytropic standard model in dynes/cm^2, see Equation (2-315) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  variable rhoc = (Pc/K)^(3./4.); % central density of a polytropic standard model in g/cm^3, polytropic equation of state for n=3
  variable Lane_Emden_function = solve_Lane_Emden_equation(3);
  variable s = struct{ r, rho, P, T };
  s.r = Lane_Emden_function.xi/Lane_Emden_function.xi[-1]*R; % radius in solar radii
  s.rho = rhoc*Lane_Emden_function.theta^3; % density in g/cm^3, see Equation (2-291) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  s.P = K*s.rho^(4./3.); % pressure in dynes/cm^2, polytropic equation of state for n=3
  s.T = (8.314510e7/mu*3/7.5646e-15*(1.-beta)/beta*s.rho)^(1./3.); % temperature in K, see Equation (2-287) in the book "Principles of Stellar Evolution and Nucleosynthesis" by D.D. Clayton
  return s;
}%}}}
