define position_angle()
%!%+
%\function{position_angle}
%\usage{ sep=position_angle(ra1,dec1,ra2,dec2);}
%\synopsis{calculates the position angle of object1 with respect to object2}
%\qualifiers{
%\qualifier{deg}{if set, the input coordinates and output are in degrees (default: radian)}
%\qualifier{radian}{if set, the input coordinates and output are in radian (the default))}
%}
%\description
% This routine calculates the position angle on the sky of the point 
% with coordinates ra1/dec1 with respect to the object with coordinates
% ra2/dec2.
%
% Following the IAU conventions, the position angle is 0 if object1 is
% directly north of 2 and it is counted in the N-E-S-W-N direction, i.e.,
% it is pi/2 / 90deg if object 1 is East of object 2.
%
% Note that per IAU convention, "position angle" always refers to the
% relative position with respect to the equatorial system. This
% routine is agnostic to the type of coordinate system, i.e., the
% position angle can be computed in all coordinate systems, but the
% term "position angle" should be avoided for other coordinate systems.
%
% The position angle is not defined if object 2 is at one of the poles,
% it is 0 if object1 and object2 are at the same position.
%
% The position angle is returned in the units of the input coordinate system.
%
% This function is array safe for either ra1/dec1 or for ra2/dec2.
%
%\seealso{hms2deg,dms2deg,angle2string,greatcircle_distance}
%!%-
{
    variable ra1in,dec1in,ra2in,dec2in;
    variable ra1,dec1,ra2,dec2;
    switch(_NARGS)
    { case 4: (ra1in,dec1in,ra2in,dec2in)=(); }
    { help(_function_name()); return; }

    if (qualifier_exists("deg")) {
	ra1=ra1in*PI/180.;
	dec1=dec1in*PI/180.;
	ra2=ra2in*PI/180.;
	dec2=dec2in*PI/180.;
    } else {
	ra1=@ra1in;
	dec1=@dec1in;
	ra2=@ra2in;
	dec2=@dec2in;
    }
    
    variable s,c,td1,s2,c2;
    (s,c)=sincos(ra1-ra2);
    td1=tan(dec1);
    (s2,c2)=sincos(dec2);
    variable pa= atan2(s,c2*td1-s2*c);

    if (typeof(pa)==Array_Type) {
        variable ndx=where(pa<0.);
        if (length(ndx)>0) {
            pa[ndx]+=2*PI;
        }
    } else {
        if (pa<0.) {
            pa+=2*PI;
        }
    }
    
    if (qualifier_exists("deg")) {
	return pa*180./PI;
    }
    return pa;
}
