%%%%%%%%%%%%%%%%%%%%%%%%
define radius_to_unit()
%%%%%%%%%%%%%%%%%%%%%%%%
%!%+
%\function{radius_to_unit}
%\synopsis{converts a radius from a given unit into another}
%\usage{Double_Type = radius_to_unit(Double_Type radius, String_Type from_unit, to_unit); }
%\qualifiers{
%  \qualifier{asini}{projected semi major axis (lt-s)}
%  \qualifier{i}{inclination (degrees)}
%  \qualifier{mq}{mass ratio of star to companion}
%}
%\description
%  Converts the 'radius' from a given unit 'from_unit'
%  into another unit 'to_unit'. There might be additional
%  information needed for the calculation, e.g. the
%  projected semi major axis, which is passed via
%  qualifiers.
%
%  The following units are supported (with additional
%  needed qualifiers):
%    rsun - solar radii
%    cm   - centimeters
%    lts  - light seconds
%    disp - binary displacement (asini, i, mq)
%    fill - fill factor = radius/critical radius (asini, i, mq)
%
%  Note:
%  It might be possible, that less qualifiers may be passed,
%  if the variables are canceled during the calculation. For
%  example, converting from binary displacement to fill factor
%  requires the mass ratio 'mq' only.
%
%  Note:
%  Let the projected semi major axis be the distance from CM
%  to M1. Then, mq is defined as M1/M2.
%  That means, to convert the radius of the companion, mq has
%  to be inverted and asini has to be the distance from CM
%  to M2, which is simply asini*mq[not inverted]
%\seealso{Roche_potential, Roche_critical}
%!%-
{
 variable radius, fromunit, tounit;
 switch(_NARGS)
   { case 3: (radius, fromunit, tounit) = (); }
   { help(_function_name()); return; }

 variable known_units = ["rsun","lts","disp","fill","cm"];
 % check given units
 if (wherefirst(known_units == fromunit) == NULL) { vmessage("%s: unit '%s' unkown!", _function_name, fromunit); return; }
 if (wherefirst(known_units == tounit) == NULL) { vmessage("%s: unit '%s' unkown!", _function_name, tounit); return; }
  
 variable rsol = 6.9599e10; % solar radius in cm (https://www.cfa.harvard.edu/)

 % determine needed qualifiers by specifying the conversion formula:
 % index corresponds to a parameter (rsol, asini, sin(i), (1+mq), rcrit, Const_c) and value to multiplication or division (1,-1)
 variable convert_from = [0,0,0,0,0,0], convert_to = [0,0,0,0,0,0];
 % convert input to lt-s
 switch(fromunit)
   { case "rsun": convert_from[[0,5]] = [1,-1]; } % *rsol/Const_c
   { case "cm": convert_from[5] = -1; } % /Const_c
   { case "disp": convert_from[[1,2,3]] = [1,-1,1]; } % *asini/sin(i)*(1+mq)
   { case "fill": convert_from[[1,2,3,4]] = [1,-1,1,1]; } % *asini/sin(i)*(1+mq)*rcrit
 % convert radius in lt-s to output unit
 switch(tounit)
   { case "rsun": convert_to[[0,5]] = [-1,1]; } % /rsol*Const_c
   { case "cm": convert_to[5] = 1; } % *Const_c
   { case "disp": convert_to[[1,2,3]] = [-1,1,-1]; } % /asini*sin(i)/(1+mq)
   { case "fill": convert_to[[1,2,3,4]] = [-1,1,-1,-1]; } % /asini*sin(i)/(1+mq)/rcrit
 % combine to get final conversion formula
 variable convert = convert_from + convert_to;

 % perform conversation and check on needed qualifiers
 variable asini, i, mq, rcrit;
 if (convert[0] != 0) % rsol
 {
   if (convert[0] == 1) radius *= rsol; else radius /= rsol;
 }
 if (convert[1] != 0) % asini
 {
   ifnot (qualifier_exists("asini")) { vmessage("%s: qualifier 'asini' needed to convert radius!", _function_name); return; }
   asini = qualifier("asini");
   if (convert[1] == 1) radius *= asini; else radius /= asini;
 }
 if (convert[2] != 0) % i
 {
   ifnot (qualifier_exists("i")) { vmessage("%s: qualifier 'i' needed to convert radius!", _function_name); return; }
   i = qualifier("i");
   if (convert[2] == 1) radius *= sin(1.*i*PI/180); else radius /= sin(1.*i*PI/180);
 }
 if (convert[3] != 0) % mq
 {
   ifnot (qualifier_exists("mq")) { vmessage("%s: qualifier 'mq' needed to convert radius!", _function_name); return; }
   mq = qualifier("mq");
   if (convert[3] == 1) radius *= (1.+mq); else radius /= (1.+mq);
 }
 if (convert[4] != 0) % rcrit
 {
   ifnot (qualifier_exists("mq")) { vmessage("%s: qualifier 'mq' needed to convert radius!", _function_name); return; }
   mq = qualifier("mq");
   rcrit = Roche_critical(mq).rcrit;
   if (convert[4] == 1) radius *= rcrit; else radius /= rcrit;
 }
 if (convert[5] != 0) % Const_c
 {
   if (convert[5] == 1) radius *= 2.99792458e10; else radius /= 2.99792458e10;
 }

 return radius;
}
