% Stolen from the HSLUV project, C implementation

private variable Triplet_m = {
  [3.24096994190452134377, -1.53738317757009345794, -0.49861076029300328366],
  [-0.96924363628087982613,  1.87596750150772066772,  0.04155505740717561247],
  [0.05563007969699360846, -0.20397695888897656435,  1.05697151424287856072]
};

private variable Triplet_minv = {
  [0.41239079926595948129,  0.35758433938387796373,  0.18048078840183428751],
  [0.21263900587151035754,  0.71516867876775592746,  0.07219231536073371500],
  [0.01933081871559185069,  0.11919477979462598791,  0.95053215224966058086]
};

private variable Ref_U = 0.19783000664283680764;
private variable Ref_V = 0.46831999493879100370;

private variable Kappa = 903.29629629629629629630;
private variable Epsilon = 0.00885645167903563082;

private define get_bounds (l)
{
  variable tl = l+16.0;
  variable sub1 = (tl*tl*tl)/1560896.0;
  variable sub2 = (sub1 > Epsilon ? sub1 : (l / Kappa));

  variable bounds = Double_Type[6,2];

  variable i,j;
  _for i (0, 2)
  {
    variable m1 = Triplet_m[i][0];
    variable m2 = Triplet_m[i][1];
    variable m3 = Triplet_m[i][2];

    _for j (0, 1)
    {
      variable top1 = (284517.0 * m1 - 94839.0 * m3) * sub2;
      variable top2 = (838422.0 * m3 + 769860.0 * m2 + 731718.0 * m1) * l * sub2 -  769860.0 * j * l;
      variable bottom = (632260.0 * m3 - 126452.0 * m2) * sub2 + 126452.0 * j;

      bounds[i*2 + j, 0] = top1/bottom;
      bounds[i*2 + j, 1] = top2/bottom;
    }
  }

  return bounds;
}

private define intersect_line_line (line1, line2)
{
  return (line1[1] - line2[1])/(line2[0] - line1[0]);
}

private define ray_length_until_intersect (theta, line)
{
  return line[1]/(sin(theta)-line[0]*cos(theta));
}

private define max_safe_chroma_for_l (l)
{
  variable min_len_sqr = DOUBLE_MAX;
  variable bounds = get_bounds(l);

  variable i;
  _for i (0, 5)
  {
    variable m1 = bounds[i, 0];
    variable b1 = bounds[i, 1];

    variable line2 = [-1.0/m1, 0.0];
    variable x = intersect_line_line (bounds[i, *], line2);
    variable dist = sumsq([x, b1+x*m1]);

    if (dist < min_len_sqr)
      min_len_sqr = dist;
  }

  return sqrt(min_len_sqr);
}

private define max_chroma_for_lh (l, h)
{
  variable min_len = DOUBLE_MAX;
  variable hrad = h * 0.01745329251994329577; % (2 * pi / 360)

  variable bounds = get_bounds(l);

  variable i;
  _for i (0, 5)
  {
    variable len = ray_length_until_intersect(hrad, bounds[i,*]);

    if (len >= 0 && len < min_len)
      min_len = len;
  }

  return min_len;
}

private define from_linear (c)
{
  if (c <= 0.0031308)
    return 12.92 * c;
  else
    return 1.055 * c^(1.0/2.4) - 0.055;
}

private define to_linear (c)
{
  if (c > 0.04045)
    return ((c + 0.055)/1.055)^2.4;
  else
    return c / 12.92;
}

private define xyz2rgb (x, y, z)
{
  variable r = from_linear(sum(Triplet_m[0]*[x,y,z]));
  variable g = from_linear(sum(Triplet_m[1]*[x,y,z]));
  variable b = from_linear(sum(Triplet_m[2]*[x,y,z]));

  return r,g,b;
}

private define rgb2xyz (r, g, b)
{
  variable rgbl = array_map(Double_Type, &to_linear, [r,g,b]);
  variable x, y, z;
  x = sum(Triplet_minv[0]*rgbl);
  y = sum(Triplet_minv[1]*rgbl);
  z = sum(Triplet_minv[2]*rgbl);

  return x, y, z;
}

private define y2l (y)
{
  if (y <= Epsilon)
    return y*Kappa;
  else
    return 116.0*y^(1./3) - 16.0;
}

private define l2y (l)
{
  if (l <= 8.0)
    return l/Kappa;
  else
    return ((l+16.0)/116.0)^3;
}

private define xyz2luv (x, y, z)
{
  variable var_u = (4.0 * x) / (x + (15.0 * y) + (3.0 * z));
  variable var_v = (9.0 * y) / (x + (15.0 * y) + (3.0 * z));

  variable l = y2l(y);
  variable u = 13.0*l*(var_u - Ref_U);
  variable v = 13.0*l*(var_v - Ref_V);

  if (l < 0.00000001)
    return l, 0.0, 0.0;
  else
    return l, u, v;
}

private define luv2xyz (l, u, v)
{
  if (l <= 0.00000001)
    return 0.0, 0.0, 0.0;

  variable var_u = u/(13.0*l) + Ref_U;
  variable var_v = v/(13.0*l) + Ref_V;
  variable y = l2y(l);
  variable x = -(9.0*y*var_u)/((var_u-4.0)*var_v-var_u*var_v);
  variable z = (9.0*y-(15.0*var_v*y)-(var_v*x))/(3.0*var_v);

  return x, y, z;
}

private define luv2lch (l, u, v)
{
  variable h, c = sqrt(u*u+v*v);

  if (c < 0.00000001)
    h = 0.0;
  else
  {
    h = atan2(v, u) * 57.29577951308232; % 180. / PI
    if (h < 0)
      h += 360.;
  }

  return l, c, h;
}

private define lch2luv (l, c, h)
{
  variable hrad = h * 0.017453292519943295; % PI / 180.
  variable u = cos(hrad)*c;
  variable v = sin(hrad)*c;

  return l, u, v;
}

private define hsluv2lch (h, s, l)
{
  variable c;

  if (l > 99.9999999 || l < 0.00000001)
    c = 0.0;
  else
    c = max_chroma_for_lh(l,h)/100.0*s;

  if (s<0.00000001)
    h = 0.0;

  return l, c, h;
}

private define lch2hsluv (l, c, h)
{
  variable s;

  if (l > 99.9999999 || l < 0.00000001)
    s = 0.0;
  else
    s = c / max_chroma_for_lh(l, h) * 100.0;

  if (c < 0.00000001)
    h = 0.0;

  return h, s, l;
}

private define hpluv2lch (h, s, l)
{
  variable c;

  if(l > 99.9999999 || l < 0.00000001)
    c = 0.0;
  else
    c = max_safe_chroma_for_l(l) / 100.0 * s;

  if (s < 0.00000001)
    h = 0.0;

  return l, c, h;
}

private define lch2hpluv (l, c, h)
{
  variable s;

  if (l > 99.9999999 || l < 0.00000001)
    s = 0.0;
  else
    s = c / max_safe_chroma_for_l(l) * 100.0;

  if (c < 0.00000001)
    h = 0.0;

  return h, s, l;
}

define hsluv2rgb (h, s, l)
%!%+
%\function{hsluv2rgb}
%\usage{Int_Type r, g, b = hsluv2rgb(Double_Type h, s, l);}
%\synopsis{Calculate RGB triplet from HSLuv space}
%\description
%  This function returns the RGB colors (as 8 bit values) of
%  the corresponding HSLuv color (see https://www.hsluv.org/).
%  The inputs range from 0 to 1.
%
%  This colorpsace is useful to generate custom color maps and
%  palettes as it attemps to give the same perceived lightness (l)
%  and saturation (s) for different hues (h). It does not contain
%  the full RGB space, however.
%
%\seealso{rgb2hsluv}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &hsluv2lch, h*360.,s*100.,l*100.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2xyz, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2rgb, A, B, C);

  (A, B, C) = (int(_max(0, _min(A*255, 255))), int(_max(0, _min(B*255, 255))), int(_max(0, _min(C*255, 255))));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define hpluv2rgb (h, s, l)
%!%+
%\function{hpluv2rgb}
%\usage{Int_Type r, g, b = hpluv2rgb(Double_Type h, s, l);}
%\synopsis{Calculate RGB triplet from HPLuv space}
%\description
%  This function returns the RGB colors (as 8 bit values) of
%  the corresponding HPLuv color (see https://www.hsluv.org/).
%  The inputs range from 0 to 1.
%
%  This colorpsace is useful to generate custom color maps and
%  palettes as it attemps to give the same perceived lightness (l)
%  and saturation (s) for different hues (h). Compared to HSLuv
%  it tries to maximize the used color range.
%
%\seealso{rgb2hpluv, hsluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &hpluv2lch, h*360.,s*100.,l*100.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2xyz, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2rgb, A, B, C);

  (A, B, C) = (int(_max(0, _min(A*255, 255))), int(_max(0, _min(B*255, 255))), int(_max(0, _min(C*255, 255))));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define rgb2hsluv (r, g, b)
%!%+
%\function{rgb2hsluv}
%\usage{Double_Type h, s, l = rgb2hsluv(Int_Type r, g, b);}
%\synopsis{Calculate HSLuv triplet from RGB space}
%\description
%  Inverse function of \code{hsluv2rgb}.
%
%\seealso{hsluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &rgb2xyz, r/255., g/255., b/255.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2lch, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2hsluv, A, B, C);

  (A, B, C) = (_max(0, _min(A/360., 1)), _max(0, _min(B/100., 1)), _max(0, _min(C/100., 1)));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}

define  rgb2hpluv (r, g, b)
%!%+
%\function{rgb2hpluv}
%\usage{Double_Type h, s, l = rgb2hpluv(Int_Type r, g, b);}
%\synopsis{Calculate HSPuv triplet from RGB space}
%\description
%  Inverse function of \code{hpluv2rgb}.
%
%\seealso{hpluv2rgb}
%!%-
{
  variable A, B, C;
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &rgb2xyz, r/255., g/255., b/255.);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &xyz2luv, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &luv2lch, A, B, C);
  (A, B, C) = array_map(Double_Type, Double_Type, Double_Type, &lch2hpluv, A, B, C);

  (A, B, C) = (_max(0, _min(A/360., 1)), _max(0, _min(B/100., 1)), _max(0, _min(C/100., 1)));

  return length(A) == 1 ? (A[0], B[0], C[0]) : (A, B, C);
}
