% -*- mode: fold; mode: slang; -*-

private variable ColorPalettes = Assoc_Type[Any_Type, [0x000000, 0xff0000, 0x00ff00, 0x0000ff, 0xffff00, 0xff00ff, 0x00ffff, 0xffffff]];
define add_color_palette (name, values)
%!%+
%\function{add_color_palette}
%\synopsis{Add a color palette}
%\usage{add_color_palette (String_Type name, UInt_Type[] colors);
%\altusage{add_color_palette (String_Type name, Ref_Type generator);}}
%\description
%  Add an array of hex values representing a color palette to the
%  system.
%
%  Alternatively, a function can be passed with this signature
%  \code{UInt_Type[] palette = generator(UInt_Type n);}
%  where \code{n} is the number of colors returned. It is okay to
%  return fewer colors than \code{n}, but never more.
%\seealso{get_color_palette}
%!%-
{
  ColorPalettes[name] = values;
}

% seaborn colors %{{{
add_color_palette("sb:deep", [0x4c72b0, 0xdd8452, 0x55a868, 0xc44e52, 0x8172b3, 0x937860,
			      0xda8bc3, 0x8c8c8c, 0xccb974, 0x64b5cd]);
add_color_palette("sb:muted", [0x4878d0, 0xee854a, 0x6acc64, 0xd65f5f, 0x956cb4, 0x8c613c,
			       0xdc7ec0, 0x797979, 0xd5bb67, 0x82c6e2]);
add_color_palette("sb:pastel", [0xa1c9f4, 0xffb482, 0x8de5a1, 0xff9f9b, 0xd0bbff, 0xdebb9b,
			       0xfab0e4, 0xcfcfcf, 0xfffea3, 0xb9f2f0]);
add_color_palette("sb:bright", [0x023eff, 0xff7c00, 0x1ac938, 0xe8000b, 0x8b2be2, 0x9f4800,
				0xf14cc1, 0xa3a3a3, 0xffc400, 0x00d7ff]);
add_color_palette("sb:dark", [0x001c7f, 0xb1400d, 0x12711c, 0x8c0800, 0x591e71, 0x592f0d,
			      0xa23582, 0x3c3c3c, 0xb8850a, 0x006374]);
add_color_palette("sb:colorblind", [0x0173b2, 0xde8f05, 0x029e73, 0xd55e00, 0xcc78bc, 0xca9161,
				    0xfbafe4, 0x949494, 0xece133, 0x56b4e9]);
%!%+
%\function{palette--sb}
%\synopsis{Get Seaborn color paletts}
%\usage{UInt_Type[] get_color_paletts("sb:"+name, UInt_Type n);}
%\description
%  The seaborn paletts are a very good set of colors for qualitative
%  representation of data (like lines, points, etc.).
%!%-

%}}}

private define hsl_palette (n)
%!%+
%\function{palette--hsl}
%\synopsis{Generate palette from HSL space}
%\usage{UInt_Type[] palette = get_color_palette ("hsl", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.65]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hsl2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsl2rgb(hues, qualifier("s", 0.65), qualifier("l", 0.6)));
}

private define hsluv_palette (n)
%!%+
%\function{palette--hsluv}
%\synopsis{Generate palette from HSLuv space}
%\usage{UInt_Type[] palette = get_color_palette ("hsluv", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.90]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hsluv2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsluv2rgb(hues, qualifier("s", 0.9), qualifier("l", 0.6)));
}

private define hpluv_palette (n)
%!%+
%\function{palette--hpluv}
%\synopsis{Generate palette from HSLuv space}
%\usage{UInt_Type[] palette = get_color_palette ("hpluv", UInt_Type n);}
%\qualifiers{
%\qualifier{h[=0.01]}{Starting hue}
%\qualifier{s[=0.90]}{Saturation}
%\qualifier{l[=0.60]}{Lightness}
%}
%\description
%  Create a color palette with \code{n} entries. Control the
%  appearence with the HSL qualifiers.
%\seealso{hpluv2rgb}
%!%-
{
  variable hues = ([0:1:#n]+qualifier("h", 0.01)) mod 1.0;
  hues -= int(hues);

  return rgb2hex(hsluv2rgb(hues, qualifier("s", 0.9), qualifier("l", 0.6)));
}

% HSL, HSLuv palettes %{{{
add_color_palette("hsl", &hsl_palette);
add_color_palette("hsluv", &hsluv_palette);
add_color_palette("hpluv", &hpluv_palette);
%}}}

% SRON (Paul Tol) palettes %{{{
private define sron_palette (n)
%!%+
%\function{palette--sron-rainbow}
%\synopsis{Get SRON rainbow palette from Paul Tol}
%\usage{UInt_Type[] palette = get_color_palette("sron-rainbow", UInt_Type n);}
%\description
%  Returns a set of colors more or less like rainbow colors with good contrast.
%  \code{n} makes a smart subset up to 23.
%!%-
{
  variable colors = [0xe8ecfb, 0xd9cce3, 0xd1bbd7, 0xcaaccb, 0xba8db4, 0xae76a3,
		     0xaa6f9e, 0x994f88, 0x882e72, 0x1965b0, 0x437dbf, 0x5289c7,
		     0x6195cf, 0x7bafde, 0x4eb265, 0x90c987, 0xcae0ab, 0xf7f056,
		     0xf7cb45, 0xf6c141, 0xf4a736, 0xf1932d, 0xee8026, 0xe8601c,
		     0xe65518, 0xdc050c, 0xa5170e, 0x72190e, 0x42150a, 0x777777];
  variable selector = {
    [10],
    [10, 26],
    [10, 18, 26],
    [10, 15, 18, 26],
    [10, 14, 15, 18, 26],
    [10, 14, 15, 17, 18, 26],
    [9, 10, 14, 15, 17, 18, 26],
    [9, 10, 14, 15, 17, 18, 23, 26],
    [9, 10, 14, 15, 17, 18, 23, 26, 28],
    [9, 10, 14, 15, 17, 18, 21, 24, 26, 28],
    [9, 10, 12, 14, 15, 17, 18, 21, 24, 26, 28],
    [3, 6, 9, 10, 12, 14, 15, 17, 18, 21, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 21, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26],
    [3, 6, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 28],
    [3, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 12, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 20, 22, 24, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28],
    [2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28, 29],
    [1, 2, 4, 5, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17, 18, 19, 21, 23, 25, 26, 27, 28, 29]};

  if (n > length(selector))
    n = length(selector);

  return colors[selector[n-1]];
}
add_color_palette("sron-rainbow", &sron_palette);

add_color_palette("sron:bright", [0x4477aa, 0x66ccee, 0x228833, 0xccbb44, 0xee6677, 0xaa3377, 0xbbbbbb]);
add_color_palette("sron:high-contrast", [0xffffff, 0xddaa33, 0xbb5566, 0x004488, 0x000000]);
add_color_palette("sron:vibrant", [0x0077bb, 0x33bbee, 0x009988, 0xee7733, 0xcc3311, 0xee3377,
				   0xbbbbbb]);
add_color_palette("sron:muted", [0x332288, 0x88ccee, 0x44aa99, 0x117733, 0x999933, 0xddcc77,
				 0xcc6677, 0x882255, 0xaa4499, 0xdddddd]);
add_color_palette("sron:medium-contrast", [0xffffff, 0xeecc66, 0xee99aa, 0x6699cc, 0x997700, 0x994455,
					   0x004488, 0x000000]);
add_color_palette("sron:pale", [0xbbccee, 0xcceeff, 0xccddaa, 0xeeeebb, 0xffcccc, 0xdddddd]);
add_color_palette("sron:dark", [0x222255, 0x225555, 0x225522, 0x666633, 0x663333, 0x555555]);
add_color_palette("sron:light", [0x77aadd, 0x99ddff, 0x44bb99, 0xbbcc33, 0xaaaa00, 0xeedd88,
				 0xee8866, 0xffaabb, 0xdddddd]);
add_color_palette("sron:sunset", [0x364b9a, 0x4a7bb7, 0x6ea6cd, 0x98cae1, 0xc2e4ef, 0xeaeccc,
				  0xfeda8b, 0xfdb366, 0xf67e4b, 0xdd3d2d, 0xa50026, 0xffffff]);
add_color_palette("sron:burd", [0x2166ac, 0x4393c3, 0x92c5de, 0xd1e5f0, 0xf7f7f7, 0xfddbc7,
				0xf4a582, 0xd6604d, 0xb2182b, 0xffee99]);
add_color_palette("sron:prgn", [0x762a83, 0x9970ab, 0xc2a5cf, 0xe7d4e8, 0xf7f7f7, 0xd9f0d3,
				0xacd39e, 0x5aae61, 0x187837, 0xffee99]);
add_color_palette("sron:ylorbr", [0xffffe5, 0xfff7bc, 0xfee391, 0xfec44f, 0xfb9a29, 0xec7014,
				  0xcc4c02, 0x993404, 0x662506, 0x888888]);
add_color_palette("sron:iridescent", [0xfefbe9, 0xfcf7d5, 0xf5f3c1, 0xeaf0b5, 0xddecbf, 0xd0e7ca,
				      0xc2e3d2, 0xb5ddd8, 0xa8d8dc, 0x9bd2e1, 0x8dcbe4, 0x81c4e7,
				      0x7bbce7, 0x7eb2e4, 0x88a5dd, 0x9398d2, 0x9b8ac4, 0x9d7db2,
				      0x9a709e, 0x906388, 0x805770, 0x684957, 0x46353a, 0x999999]);
add_color_palette("sron:smooth-rainbow", [0xe8ecfb, 0xddd8ef, 0xd1c1e1, 0xc3a8d1, 0xb58fc2, 0xa778b4,
					  0x9b62a7, 0x8c4e99, 0x6f4c9b, 0x6059a9, 0x5568b8, 0x4e79c5,
					  0x4d8ac6, 0x4e96bc, 0x549eb3, 0x59a5a9, 0x60ab9e, 0x69b190,
					  0x77b77d, 0x8cbc68, 0xa6be54, 0xbebc48, 0xd1b541, 0xddaa3c,
					  0xe49c39, 0xe78c35, 0xe67932, 0xe4632d, 0xdf4828, 0xda2222,
					  0xb8221e, 0x95211b, 0x721e17, 0x521a13, 0x666666]);
add_color_palette("sron:landcover", [0x5566aa, 0x117733, 0x44aa66, 0x55aa22, 0x668822, 0x99bb55,
				     0x558877, 0x88bbaa, 0xaaddcc, 0x44aa88, 0xddcc66, 0xffdd44,
				     0xffee88, 0xbb0011]);
%!%+
%\function{palette--sron}
%\synopsis{Get SRON palettes}
%\usage{UInt_Type[] get_color_palette("sron:"+name, UInt_Type n);}
%\description
%  The SRON colors are defined in Paul Tol's technical note (\code{https://personal.sron.nl/~pault/data/colourschemes.pdf})
%  and provides a set of colors for qualitative, diverging, and sequential
%  representations.
%
%  Note: To get the original set of colors (was provided by \code{get_sron_colors})
%  use \code{get_color_palette("sron-rainbow", n);} for diverging and sequential
%  there are no direct matches. See the other color palette options.
%
%\seealso{get_color_palette_names, palette--sron-rainbow}
%!%-
%}}}

% Tableau10 palettes %{{{
add_color_palette("tab:10", [0x4e79a7, 0xf28e2b, 0xe15759, 0x76b7b2, 0x59a14f, 0xedc948,
			    0xb07aa1, 0xff9da7, 0x9c755f, 0xbab0ac]);
add_color_palette("tab:20", [0x4e79a7, 0xa0cbe8, 0xf28e2b, 0xffbe7d, 0x59a14f, 0x8cd17d,
			    0xb6992d, 0xf1ce63, 0x499894, 0x86bcb6, 0xe15759, 0xff9d9a,
			    0x79706e, 0xbab0ac, 0xd37295, 0xfabfd2, 0xb07aa1, 0xd4a6c8,
			    0x9d7660, 0xd7b5a6]);
add_color_palette("tab:colorblind", [0x1170aa, 0xfc7d0b, 0xa3acb9, 0x57606c, 0x5fa2ce, 0xc85200,
				     0x7b848f, 0xa3cce9, 0xffbc79, 0xc8d0d9]);
add_color_palette("tab:seattle-grays", [0x767f8b, 0xb3b7b8, 0x5c6060, 0xd3d3d3, 0x989ca3]);
add_color_palette("tab:traffic", [0xb60a1c, 0xe39802, 0x309143, 0xe03531, 0xf0bd27, 0x51b364,
				  0xff684c, 0xffda66, 0x8ace7e]);
add_color_palette("tab:miller-stone", [0x4f6980, 0x849db1, 0xa2ceaa, 0x638b66, 0xbfbb60, 0xf47942,
				       0xfbb04e, 0xb66353, 0xd7ce9f, 0xb9aa97, 0x7e756d]);
add_color_palette("tab:superfishel-stone", [0x6388b4, 0xffae34, 0xef6f6a, 0x8cc2ca, 0x55ad89, 0xc3bc3f,
					    0xbb7693, 0xbaa094, 0xa9b5ae, 0x767676]);
add_color_palette("tab:nuriel-stone", [0x8175aa, 0x6fb899, 0x31a1b3, 0xccb22b, 0xa39fc9, 0x94d0c0,
				       0x959c9e, 0x027b8e, 0x9f8f12]);
add_color_palette("tab:jewel-bright", [0xeb1e2c, 0xfd6f30, 0xf9a729, 0xf9d23c, 0x5fbb68, 0x64cdcc,
				       0x91dcea, 0xa4a4d5, 0xbbc9e5]);
add_color_palette("tab:summer", [0xbfb202, 0xb9ca5d, 0xcf3e53, 0xf1788d, 0x00a2b3, 0x97cfd0,
				 0xf3a546, 0xf7c480]);
add_color_palette("tab:winter", [0x90728f, 0xb9a0b4, 0x9d983d, 0xcecb76, 0xe15759, 0xff9888,
				 0x6b6b6b, 0xbab2ae, 0xaa8780, 0xdab6af]);
add_color_palette("tab:gree-orange-teal", [0x4e9f50, 0x87d180, 0xef8a0c, 0xfcc66d, 0x3ca8bc, 0x98d9e4,
					   0x94a323, 0xc3ce3d, 0xa08400, 0xf7d42a, 0x26897e, 0x8dbfa8]);
add_color_palette("tab:red-blue-brown", [0x466f9d, 0x91b3d7, 0xed444a, 0xfeb5a2, 0x9d7660, 0xd7b5a6,
					 0x3896c4, 0xa0d4ee, 0xba7e45, 0x39b87f, 0xc8133b, 0xea8783]);
add_color_palette("tab:purple-pink-gray", [0x8074a8, 0xc6c1f0, 0xc46487, 0xffbed1, 0x9c9290, 0xc5bfbe,
					   0x9b93c9, 0xddb5d5, 0x7c7270, 0xf498b6, 0xb173a0, 0xc799bc]);
add_color_palette("tab:hue-circle", [0x1ba3c6, 0x2cb5c0, 0x30bcad, 0x21b087, 0x33a65c, 0x57a337,
				     0xa2b627, 0xd5bb21, 0xf8b620, 0xf89217, 0xf06719, 0xe03426,
				     0xf64971, 0xfc719e, 0xeb73b3, 0xce69be, 0xa26dc2, 0x7873c0,
				     0x4f7cba]);
%!%+
%\function{palette--tab}
%\synopsis{Get Tableau10 color palettes}
%\usage{UInt_Type[] palette = get_color_palette("tab:"+name, UInt_Type n);}
%\description
%  Color palettes used by the Tableau10 software for visualization. Usuall only
%  good for qualitative data.
%!%-
%}}}

% Seaborn palette generators %{{{
private define space_converter (r, g, b)
{
  variable space = qualifier("space", "rgb");
  variable back = NULL;

  if (typeof(space) == String_Type)
  {
    if (space == "rgb")
    {
      (r,g,b) = (r/255., g/255., b/255.);
    }
    else if (space == "hsl")
    {
      (r, g, b) = rgb2hsl(r, g, b);
      back = &hsl2rgb;
    }
    else if (space == "hsluv")
    {
      (r, g, b) = rgb2hsluv(r, g, b);
      back = &hsluv2rgb;
    }
    else if (space == "hpluv")
    {
      (r, g, b) = rgb2hpluv(r, g, b);
      back = &hpluv2rgb;
    }
    else
      throw UsageError, "Color space '$space' is not recognized"$;
  }
  else if (typeof(space) == List_Type and length(space)==2
	   and typeof(space[0]) == Ref_Type and typeof(space[1]) == Ref_Type)
  {
    (r, g, b) = (@space[0])(r, g, b);
    back = space[1];
  }
  else
    throw UsageError, "Unable to perform color space conversion with given functions.";

  return (r,g,b,back);
}

private define blend_palette (n)
%!%+
%\function{palette--blend}
%\synopsis{Blend one or more colors linearly (in RGB space).}
%\usage{UInt_Type get_color_palette("blend", UInt_Type n);}
%\qualifiers{
%  \qualifier{colors[=[red,gray,blue]]}{The colors to interpolate, must be given as HEX}
%  \qualifier{set[=[0:1:#length(colors)]]}{The setpoints of the colors. Can be unsorted.}
%  \qualifier{space[="hsluv"]}{The color space in which interpolation is done.}
%}
%\description
%  Blend colors to a palette from given colors. The setpoints can be
%  used to shift the color at the same position to higher or lower
%  interpolation values.
%
%  The qualifier \code{space} allows to control in what space the interpolation
%  is done. This either must be one of the known color spaces "hsl", "hsluv", 
%  "hpluv", or "rgb", or a tuple (list with 2 elements) with references
%  to two functions. The first must take three arguments (R,G,B) as 8 bit
%  encoded colors and convert it to the color space variables (x,y,z). The
%  second converts back (x,y,z) -> (R,G,B).
%!%-
{
  variable colors = qualifier("colors", [0x2525a5, 0xa5a5a5, 0xa52525]);
  variable weights = _max(0., _min(1.,
			qualifier("set", [0:1:#length(colors)])
		     ));
  if (length(colors) != length(weights))
    throw UsageError, sprintf("Length of colors (%d) and length of setpoints (%d) do not match",
			      length(colors), length(weights));
  variable sort = array_sort(weights);
  weights = weights[sort];
  colors = colors[sort];

  variable r,g,b;
  (r,g,b) = hex2rgb(colors);

  variable back;
  (r, g, b, back) = space_converter(r, g, b; space=qualifier("space", "hsluv"));

  % linear interpolation of the color space
  variable dw = weights[[1:]]-weights[[:-2]];
  variable mr = (r[[1:]]-r[[:-2]])*1./dw;
  variable mg = (g[[1:]]-g[[:-2]])*1./dw;
  variable mb = (b[[1:]]-b[[:-2]])*1./dw;
  variable winf = where(dw==0);
  mr[winf] = 0.;
  mg[winf] = 0.;
  mb[winf] = 0.;

  weights = weights[[:-2]];
  variable idx = [0, int(weights*(n-1)), n-1];
  variable tr = r[[:-2]]-mr*weights;
  variable tg = g[[:-2]]-mg*weights;
  variable tb = b[[:-2]]-mb*weights;
  variable grid = [0:1:#n];
  r = Double_Type[n];
  g = Double_Type[n];
  b = Double_Type[n];
  variable w, i, m, t;

  w = [0:idx[0]-1];
  r[w] = _max(0., _min(grid[w]*mr[0]+tr[0], 1.));
  g[w] = _max(0., _min(grid[w]*mg[0]+tg[0], 1.));
  b[w] = _max(0., _min(grid[w]*mb[0]+tb[0], 1.));
  _for i (1, length(idx)-3)
  {
    w = [idx[i]:idx[i+1]-1];
    r[w] = _max(0., _min(grid[w]*mr[i-1]+tr[i-1], 1.));
    g[w] = _max(0., _min(grid[w]*mg[i-1]+tg[i-1], 1.));
    b[w] = _max(0., _min(grid[w]*mb[i-1]+tb[i-1], 1.));
  }
  w = [idx[-2]:idx[-1]];
  r[w] = _max(0., _min(grid[w]*mr[-1]+tr[-1], 1.));
  g[w] = _max(0., _min(grid[w]*mg[-1]+tg[-1], 1.));
  b[w] = _max(0., _min(grid[w]*mb[-1]+tb[-1], 1.));

  if (back == NULL)
    return rgb2hex(r, g, b);
  else
    return rgb2hex(@back(r, g, b));
}
add_color_palette("blend", &blend_palette);

private define dark_palette (n)
%!%+
%\function{palette--dark}
%\synopsis{Get dark blend palette with given color}
%\usage{UInt_Type[] palette = get_color_palette("dark", UInt_Type n);}
%\qualifiers{
%  \qualifier{color[=red]}{Color to blend with dark pattern (as HEX value).}
%  \qualifier{reverse}{If given, blend from color to dark}
%}
%\description
%  Get a color palette from the chosen color fading in from
%  a dark tone.
%!%-
{
  variable color = qualifier("color", 0xa52525);
  variable r, g, b;
  (r, g, b) = hex2rgb(color);
  variable h,s;
  (h,s,) = rgb2hsluv(r, g, b);
  variable gray = rgb2hex(hsluv2rgb(h, 0.15*s, 0.15));

  variable colors;
  if (qualifier_exists("reverse"))
    colors = [rgb2hex(r, g, b), gray];
  else
    colors = [gray, rgb2hex(r, g, b)];

  return blend_palette(n;; struct { @(__qualifiers()), colors=colors});
}
add_color_palette("dark", &dark_palette);

private define light_palette (n)
%!%+
%\function{palette--light}
%\synopsis{Get light blend palette with given color}
%\usage{UInt_Type[] palette = get_color_palette("light", UInt_Type n);}
%\qualifiers{
%  \qualifier{color[=red]}{Color to blend with dark pattern (as HEX value).}
%  \qualifier{reverse}{If given, blend from color to light}
%}
%\description
%  Get a color palette from the chosen color fading in from
%  a light tone.
%!%-
{
  variable color = qualifier("color", 0xa52525);
  variable r, g, b;
  (r, g, b) = hex2rgb(color);
  variable h,s;
  (h,s,) = rgb2hsluv(r, g, b);
  variable gray = rgb2hex(hsluv2rgb(h, 0.15*s, 0.95));

  variable colors;
  if (qualifier_exists("reverse"))
    colors = [rgb2hex(r, g, b), gray];
  else
    colors = [gray, rgb2hex(r, g, b)];

  return blend_palette(n;; struct { @(__qualifiers()), colors=colors});
}
add_color_palette("light", &light_palette);

private define diverging_palette (n)
%!%+
%\function{palette--diverging}
%\synopsis{Create a diverging palette between two colors}
%\usage{UInt_Type palette = get_color_palette("diverging", UInt_Type n);}
%\qualifiers{
%  \qualifier{dark}{If given, midpoint is a dark tone}
%  \qualifier{h_pos[=0.7]}{Hue of positive color}
%  \qualifier{h_neg[=0.2]}{Hue of negative color}
%  \qualifier{sep[=0.01]}{Size of midpoint region}
%  \qualifier{s[=0.75]}{Saturation of both colors}
%  \qualifier{l[=0.5]}{Luminosity of both colors}
%}
%\description
%  Create a color palette useful to represent diverging data.
%
%\seealso{hsluv2rgb}
%!%-
{
  variable blender = (qualifier_exists("dark")) ?
    &dark_palette: &light_palette;
  variable mid = (qualifier_exists("dark")) ?
    rgb2hex(hsluv2rgb(0.133, 0.133, 0.133)) : rgb2hex(hsluv2rgb(0.95, 0.95, 0.95));

  variable sep = _max(0.0, _min(1.0, qualifier("sep", 0.01)));
  variable s = qualifier("s", 0.75);
  variable l = qualifier("l", 0.5);
  variable neg = rgb2hex(hsluv2rgb(qualifier("h_neg", 0.611), s, l));
  variable pos = rgb2hex(hsluv2rgb(qualifier("h_pos", 0.055), s, l));
  variable sepw = int(sep*255);
  variable half = int(128-sepw/2);

  neg = (@blender)(half; color=neg, reverse);
  pos = (@blender)(half; color=pos);

  return blend_palette(n; colors=[neg, Int_Type[sepw]+mid, pos]);
}
add_color_palette("diverging", &diverging_palette);

private define cubehelix_palette (n)
%!%+
%\function{palette--cubehelix}
%\synopsis{Create palette using 'cubehelix' description}
%\usage{UInt_Type[] pattern = get_color_palette("cubehelix", UInt_Type n);}
%\qualifiers{
%  \qualifier{start[=0.0]}{Start color of the palette}
%  \qualifier{s[=0.8]}{Saturation of the colors}
%  \qualifier{rot[=0.4]}{How far the scale rotates in color (hue)}
%  \qualifier{dark[=0.15]}{Dark end of the palette}
%  \qualifier{light[=0.85]}{Light end of the palette}
%}
%\description
%  Calculates a color palette according to the description
%  in D. A. Green's, 'A colour scheme for the display of
%  astronomical intensity images', (Bull. Astr. Soc. India (2011)).
%!%-
{
  variable _r = [-0.14861, 1.78277];
  variable _g = [-0.29227, -0.90649];
  variable _b = [1.97294, 0.0];

  variable r, g, b;
  variable start = _max(0., _min(1., qualifier("start", 0.0)));
  variable s = _max(0.0, _min(1., qualifier("s", 0.8)));
  variable rot = qualifier("rot", 0.4);
  variable gamma = qualifier("gamma", 1.0);
  variable grid = [_max(0.0, _min(1.0, qualifier("dark", 0.15))):
		   _max(0.0, _min(qualifier("light", 0.85))):#n];
  variable lambda = grid^gamma;

  variable phi = 2*PI*(start + rot*grid);
  variable a = s*lambda*(1.0-lambda)/2.0;
  variable sphi = sin(phi);
  variable cphi = cos(phi);

  r = _max(0., _min(1., lambda + a*(_r[0]*sphi+_r[1]*cphi)));
  g = _max(0., _min(1., lambda + a*(_g[0]*sphi+_g[1]*cphi)));
  b = _max(0., _min(1., lambda + a*(_b[0]*sphi+_b[1]*cphi)));

  return rgb2hex(r, g, b);
}
add_color_palette("cubehelix", &cubehelix_palette);
%}}}

private define colormap_palette (n, name)
%!%+
%\function{palette--map}
%\synopsis{Retrieve a set of colors from a colormap}
%\usage{UInt_Type[] palette = get_color_palette (String_Type name, UInt_Type len);}
%\qualifiers{
%  \qualifier{start[=0.0]}{Starting "color" (lowest value)}
%  \qualifier{range[=1.0]}{Range of "color", larger than one causes repetitions}
%}
%\description
%  This function is useful if a range of colors is required.
%  It returns \code{len} values from a linear mapping of the
%  given colormap. The colors are constructed from palette--blend
%  and forwards all qualifiers to this constructor.
%
%  Note: This function is used only if the given \code{name} is
%  a colormap (and does not match a palette name, see \code{png_get_colormap_names}),
%  to construct a %  set of colors from a color array see
%  palette--blend.
%
%\seealso{png_get_colormap_names, palette--name}
%!%-
{
#ifexists png_add_colormap
  variable map;
  try
    map = png_get_colormap(name);
  catch AnyError:
    throw UsageError, "Unknown palette: '$name'"$;

  variable start = qualifier("start", 0.0);
  variable range = qualifier("range", 1.0);
  while (start<0) start += 1.0;

  start = int(start*(length(map)-1));
  range = int(range*(length(map)-1));

  variable ind = [start:start+range] mod length(map);

  return blend_palette(n;; struct { @__qualifiers(),  colors=map[ind] });
#else
  return ColorPalettes[""];
#endif
}

define get_color_palette (name, n)
%!%+
%\function{get_color_palette}
%\synopsis{Get a number of colors from a color palette}
%\usage{UInt_Type[] = palette get_color_palette(String_Type name, UInt_Type n);}
%\qualifiers{
%  \qualifier{repeat}{If given, repeat colors to match \code{n}}
%  \qualifier{cmap}{If given, force colormap search}
%  \qualifier{string}{Return colormap as HEX string instead of value}
%}
%\description
%  Get \code{n} colors (as hex encoded values) from a palette. Known
%  palette names can be obtained by \code{get_color_palette_names}.
%
%  This function forwards all qualifiers to the generator functions (if any).
%  If no palette matching \code{name} is found, a colormap is searched.
%  To generate a palette from a colormap only, use the \code{cmap}
%  qualifier. For information how palettes are generated from colormaps
%  see \code{palette--map}, else see \code{palette--name}, where \code{name}
%  is the first argument.
%  
%  If it is not a colormap nor a generator function, a static color map is
%  used. They are usually named by like \code{family:palette}.
%  For most of them, information can be obtained via \code{palette--family}.
%
%  Eventually the used map only has a finite number of colors. If \code{n}
%  exceeds this number, only the maximum number of colors are returned. To
%  repeat colors up to \code{n} times, use the \code{repeat} qualifier.  
%
%\seealso{palette--hsl,palette--map,palette--sron}
%!%-
{
  variable pal;

  if (assoc_key_exists(ColorPalettes, name) and not qualifier_exists("cmap"))
    pal = ColorPalettes[name];
  else
    pal = colormap_palette(n, name;; __qualifiers());

  if (typeof(pal) == Ref_Type)
    pal = @pal(n;; __qualifiers());

  if (qualifier_exists("repeat") and length(pal)<n)
    pal = pal[[0:n-1] mod length(pal)];

  if (qualifier_exists("string"))
    return array_map(String_Type, &sprintf, "%06x", pal);

  return pal;
}

define get_color_palette_names ()
%!%+
%\function{get_color_palette_names}
%\synopsis{Get list of available color palette names}
%\usage{String_Type[] names = get_color_palette_names();}
%\seealso{get_color_palette}
%!%-
{
  return assoc_get_keys(ColorPalettes);
}
