define JDofDate()
%!%+
%\function{JDofDate}
%\synopsis{Calculates the (modified) Julian Date}
%\usage{Double_Type JD = JDofDate( struct { year, month, day, hour, minute, second } );
%\altusage{Double_Type JD = JDofDate(year, month, day[, hour[, minute[, second]]]);}
%}
%\qualifiers{
%\qualifier{mjd}{return the MJD, i.e., JD-2400000.5}
%\qualifier{julian_calendar}{return the JD/MJD assuming the
%       Julian calendar rather than the Gegorian one}
%}
%\description
%    This routine calculates the Julian Date for a given Gregorian date
%    following the routine by Fliegel and van Flandern (1968, Comm. ACM 11, 657)
%    and further optimized as discussed by Pulkkinen and van Flandern (1979,
%    ApJ Suppl 41: 391). With the julian_calendar qualifier the date is
%    calculated for the Julian calendar (should be done for dates before
%    15 October 1582 in Germany, see
%    https://en.wikipedia.org/wiki/Gregorian_calendar
%    for a rough table and
%    https://de.wikipedia.org/wiki/Gregorianischer_Kalender
%    for a table containing the exact dates for the switch over).
%
%    The date is either given by a structure or by at least 3 integer arguments.
%    Fractional parts for the year and month are ignored. The default values for
%    hour, minute and second are 0, fractional days, hours, etc. are properly
%    taken into account.
%
%    The year must be larger than -4716 (4717 BC).
%    This function is array safe.
%
%\seealso{MJDofDate,dateOfMJD}
%!%-
{
  variable year, month, day, hour=0, minute=0, second=0.;
  switch(_NARGS)
  { case 1:
      variable date = ();
      if(typeof(date)==Struct_Type)
      { year   = date.year;
        month  = date.month;
        day    = date.day;
        hour   = date.hour;
        minute = date.minute;
        second = date.second;
      }
      else
      { help(_function_name());
        return;
      }
  }
  { case 3: (year, month, day) = (); }
  { case 4: (year, month, day, hour) = (); }
  { case 5: (year, month, day, hour, minute) = (); }
  { case 6: (year, month, day, hour, minute, second) = (); }
  { help(_function_name()); return; }

  variable Y=typecast(year,Integer_Type);
  variable M=typecast(month,Integer_Type);
  variable D=day;

  variable jd;
  if (qualifier_exists("julian_calendar")) {
      if (M<3) {
	  Y+=-1;
	  M+=12;
      }
      variable A=int(Y/100);
      jd=int(365.25*(Y+4716))+int(30.6001*(M+1));
      jd+=( (qualifier_exists("mjd")) ? -2401525.5 : -1524 );
  } else {
      %jw do not even THINK about modifying the following lines.
      %jw the expression HAS to be done in INTEGER arithmetic, so we
      %jw better force this here
      jd=367*Y-7*(Y+(M+9)/12)/4-3*((Y+(M-9)/7)/100+1)/4+275*M/9;
      jd+=(  (qualifier_exists("mjd")) ? -678971.5 : 1721029);
  }
  
  % do this last to minimize round off errors
  jd=jd-0.5+D+(hour+(minute+second/60.)/60.)/24.;
  
  return jd;
}
